#!/usr/bin/perl

# tests for a completed Website

### TODO:
# - build restricted-access pages as public, member, and admin
# - check for valid parent pages, menu loops

use lib ".";
use strict;
use Test::More 'no_plan';
use ExSite::Config;
use ExSite::Misc;
use ExSite::Form;
use ExSite::Section;
use ExSite::Module;
use ExSite::Page;

my $section_id = shift @ARGV;

# fake Apache env
my $cwd = `pwd`;
chomp $cwd;
my @cwd = split /\//,$cwd;
$ENV{SCRIPT_NAME} = "/cgi/home.cgi";
$ENV{SERVER_NAME} = "www.foo.com";
$ENV{SCRIPT_FILENAME} = "$cwd/home.cgi";
$ENV{REQUEST_URI} = "/cgi/home.cgi";
pop @cwd;
push @cwd, "html";
$ENV{DOCUMENT_ROOT} = join "/",@cwd;

&exsite_init;
(tied %cookie)->publish_mode();
my $db = new ExSite::Form;
my (%siteurl,%sitepath);
my (%page,%servpage,%filename,%suffix,@version);

my @version = &ExSite::Config::get_versions(1);
my @site = $section_id ? 
    ( scalar $db->fetch("section",$section_id) ) :
    $db->fetch_all("section");
foreach my $s (@site) {
    next if ($s->{type} =~ /template/);
    print "\n===> $s->{title}\n\n";
    &test_section($s);
}

sub test_section {
    my $data = shift;
    my $s = new ExSite::Section(section=>$data);
    my $sid = $s->id();
    my $name = "section $sid (".$data->{title}.")";
    my ($url,$path);
    if ($data->{type} =~ /standalone/) {
	ok($data->{url}=~/https?:\/\//,"URL for standalone $name");
	ok($data->{publish_to}=~/^\//,"absolute publish_to for standalone $name");
    }
    elsif ($data->{type} =~ /native/) {
	ok(! $data->{url},"no URL for native $name");
	ok($data->{publish_to}=~/^([\w\.\-]*)$/,"relative publish_to for native $name");
    }
    my $url = $s->get_url;
    ok (! exists $siteurl{$url},"unique url for $name");
    $siteurl{$url} = $data;
    my $path = $s->diskpath;
    ok (! exists $sitepath{$path},"unique location for $name");
    $sitepath{$path} = $data;

    # fetch service pages
    my @licence = $db->fetch_match("service",{section_id=>$s->id});
    my @serv;
    foreach my $l (@licence) {
	push @serv, $l if ($l->{page_id});
    }
    %servpage = &keywise("page_id",\@serv);
    delete $servpage{0}; # module license only

    # test pages
    %page = ();
    %filename = ();
    %suffix = ();
    my @page = $s->get_all_pages();
    my %page = &keywise("page_id",\@page);
    foreach my $p (@page) {
	&test_page($p);
    }

    # any service pages left untested??
    my @servid = keys %servpage;
    ok(scalar @servid == 0,"no unused service pages ".join(",",@servid));

    # check suffixes
    ok(scalar %suffix == 1,"$name uses consistent filename suffix");

    # test templates
    my @templ = $db->fetch_match("page",{section_id=>$s->id,type=>"template"});
    foreach my $t (@templ) {
	&test_template($t);
    }

    # test libraries
    my @lib = $db->fetch_match("page",{section_id=>$s->id,type=>"library"});
    foreach my $l (@lib) {
	&test_library($l);
    }

    # translations
    &test_translations(@page);

    # posts
    &test_posts();

    #&test_zines($s);
    &test_plugins($s);
}

sub test_page {
    my $p = shift;
    $share{Page} = $p;
    delete $share{diagnostics};
    my $data = $p->get();
    return if ($data->{type} !~ /page|alternate/);
    $share{Page} = $p;
    my $pid = $data->{page_id};
    my $filename = $data->{filename};
    my $name = "page $pid ($filename)";
    ok (! exists $filename{$filename},"unique filename for $name");
    $filename{$filename} = 1;
    my $body = $p->find_content("body");
    ok(defined $body,"$name has a body");
    my $c = new ExSite::Content(content=>$body);
    my $bodyhtml = $c->get_html();
    ok($bodyhtml,"$name body has HTML content");

    if ($filename =~ /^index.html?$/) {
	# homepage!
	is($data->{publish_method},"static","$name is static");
    }

    ok($filename !~ /test/i,"$name is not a test page");

    my ($fname,$suffix) = split /\./,$filename,2;
    $suffix{$suffix} = 1;

    # template
    ok($data->{template_id},"$name has template_id set");
    return if (! $data->{template_id});
    my $t = new ExSite::Page(id=>$data->{template_id});
    is($t->get_my("type"),"template","$name has a proper template");
    return if ($t->get_my("type") ne "template");

    my $tid = $t->id();
    my %template;
    while ($tid) {
	if (exists $template{$tid}) {
	    ok(0,"no template loop on $name");
	    last;
	}
	$template{$tid} = 1;
	$t = new ExSite::Page(id=>$tid);
	$tid = $t->get_my("template_id");
    }
   
    # EXAMINE HTML
    $p->set_context();
    my $html = $p->expand();
    &test_html($html,$name,$data);

    if ($servpage{$pid}) {
	is($data->{publish_method},"dynamic","service page for $servpage{$pid}{name} is dynamic");
	if (defined $body) {
	    $bodyhtml =~ /<!--(\&+)(\w+)\((.*)\)-->/;
	    is($1,"&","non-AJAX service oyster on $name");
	    is($2,$servpage{$pid}{name},"body oyster matches service on $name");
	    is($3,"","service oyster has blank params on $name");
	}
	delete $servpage{$pid};
    }

    # error check
    if (&AtError(0)) {
	fail("no errors logged on $name");
	print &ExSite::Config::show_diagnostics("text",3),"\n";
    }
}

sub test_html {
    my ($html,$name,$data) = @_;
    ok($html,"$name has HTML");
    return if (! $html);
    ok($html =~ /<!DOCTYPE HTML/i,"HTML doctype on $name");
    ok($html =~ /<title>.+<\/title>/i,"$name has a title");
    my $i = 0;
    while ($html =~ /dcd\.cgi/) {
	$i++;
	$html =~ s/dcd\.cgi/DCD-CGI/;
    }
    if ($i > 0) {
	is($data->{publish_method},"static","AJAX plugins on static pages only; $name");
	ok($i < 3,"no more than 2 AJAX plugins on one page; $name");
    }
    $i = 0;
    while ($html =~ /content\.cgi/) {
	$i++;
	$html =~ s/content\.cgi/CONTENT-CGI/;
    }
    ok($i==0,"no unpublished content on $name");
    $i = 0;
    while ($html =~ /getdata\.cgi/) {
	$i++;
	$html =~ s/getdata\.cgi/GETDATA-CGI/;
    }
    ok($i==0,"no dynamic data views on $name");
    # mailto:
    ok($html !~ /href=.?mailto:\w/i,"no mailto: links on $name");
}

sub test_template {
    my $data = shift;
    delete $share{diagnostics};
    my $t = new ExSite::Page(data=>$data);
    my $tid = $data->{page_id};
    my $filename = $data->{filename};
    my $name = "template $tid ($filename)";
    ok (! exists $filename{"template $filename"},"unique filename for $name");
    $filename{"template $filename"} = 1;
    ok($filename !~ /test/i,"$name is not a test template");

    my %template;
    while ($tid) {
	if (exists $template{$tid}) {
	    fail("no template loop on $name");
	    last;
	}
	$template{$tid} = 1;
	$t = new ExSite::Page(id=>$tid);
	$tid = $t->get_my("template_id");
    }

    my @contents = $db->fetch_child("content","page",$data->{page_id});
    ok(scalar @contents > 0,"$name has content");
    my $ncont = scalar @contents;
    if ($ncont) {
	my %content = &keywise("name",\@contents);
	if (! $data->{template_id}) {
	    ok(exists $content{page},"base $name has page object");
	}
	my $nedit = 0;
	foreach my $c (@contents) {
	    $c++ if ($c->{type} eq "editorial");
	}
	ok($nedit < 8,"$name has fewer than 8 editorial objects");
	if ($nedit > 3) {
	    ok($nedit / $ncont < 0.5,"fewer than half of content objects in $name are editorial");
	}
    }

    # EXAMINE HTML
    $t->set_context();
    my $html = $t->expand();
    &test_html($html,$name,$data);
    # error check
    if (&AtError(0)) {
	fail("no errors logged on $name");
	print &ExSite::Config::show_diagnostics("text",3),"\n";
    }
}

sub test_library {
    my $data = shift;
    my $l = new ExSite::Page(data=>$data);
    my $lid = $data->{page_id};
    my $filename = $data->{filename};
    my $name = "library $lid ($filename)";
    ok($data->{title},"$name has a title");
    ok(! exists $filename{"library $filename"},"unique filename for $name");
    $filename{"library $filename"} = 1;
    ok($filename !~ /test/i,"$name is not a test library");
    my @contents = $db->fetch_child("content","page",$lid);
    ok(scalar @contents > 0,"$name has content");
}

sub test_translations {
    my @page = @_;
    # test translations
    my (%alt,%vers,%page);
    foreach my $p (@page) {
	next if ($p->get_my("type") !~ /page|alternate/);
	$page{$p->id} = $p;
	my $version = $p->get_my("version") || $config{default_version};
	if ($p->get_my("type") eq "alternate") {
	    $alt{$version}{$p->id} = $p;
	}
	$vers{$version}++;
    }
    ok(scalar @version == scalar keys %vers,"languages in use match defined languages");
    foreach my $v (@version) {
	next if ($v eq $config{default_version});
	foreach my $p (values %{$alt{$v}}) {
	    my $filename = $p->get_my("filename");
	    my $name = "alternate ".$p->id." ($filename)";
	    my $master = $page{$p->get_my("parent_id")};
	    ok($master,"$name has master in same section");
	    next if (! $master);
	    ok(! $master->get_my("version"),"master is $config{default_version}");
	    my @content = $db->fetch_child("content","page",$p->id);
	    ok(scalar @content > 0,"$name has content objects");
	    foreach my $c (@content) {
		my @rev = $db->fetch_child("content_data","content",$c->{content_id});
		ok(scalar @rev > 0,"$name $c->{name} has revisions");
	    }
	}
    }
    foreach my $p (values %{$alt{$config{default_version}}}) {
	foreach my $v (@version) {
	    next if ($v eq $config{default_version});
#	    %alt = &keywise("parent_id",values %{$alt{$v}});
	    ok(exists $alt{$v}{$p->id},"page ".$p->id." has $v version");
	}
    }
}    

sub test_posts {
    if (-e "test_post.csv") {
	open F, "test_post.csv";
	foreach my $data (<F>) {
	    chomp;
	    my ($page_id,$login,$path,$query,$postdata) = split /,/;
	    delete $share{identity};
	    delete $share{diagnostics};
	    delete $ENV{PATH_INFO};
	    $ENV{PATH_INFO} = $path if $path;
	    delete $ENV{QUERY_STRING};
	    $ENV{QUERY_STRING} = $query if $query;
	    $ENV{CONTENT_LENGTH} = length $postdata;
	    open POST, ">post.dat";
	    print POST $postdata;
	    close POST;
	    $ENV{CGI_INPUT_FILE} = "post.dat";
	    $ENV{REQUEST_METHOD} = "POST";

	    {
		my $p = new ExSite::Page(id=>$page_id);
		$db->do_login($login) if $login;
		
		my $pid = $data->{page_id};
		my $filename = $data->{filename};
		my $name = "page $pid ($filename)";

		# generate page
		$p->set_context();
		my $html = $p->expand();
		
		# check for errors
		if ($@) {
		    fail("perl error posting to $name: $@");
		}
		if ($!) {
		    fail("system error posting to $name: $!");
		}
		if (&ExSite::Config::AtError(0)) {
		    fail("ExSite error posting to $name:");
		    print ExSite::Config::show_diagnostics("text",3);
		}
		$@ = undef;
		$! = undef;
	    }
	}
    }
    unlink "post.dat";
}

sub test_plugins {
    my ($section) = @_;
    my $sid = $section->id;
    my @mod = &ExSite::Module::module_list();
    foreach my $modname (@mod) {
	delete $share{diagnostics};
	my $m = &ExSite::Module::get_module($modname,undef,$section);
	if (&AtError(0)) {
	    fail("no errors logged in $modname");
	    print &ExSite::Config::show_diagnostics("text",3),"\n";
	}
	ok(ref $m,"launch $modname");
	if (ref $m) {
	    if ($m->ioctl("isService")) {
		# look for appropriate service pages
		my @service = $db->custom_query("select s.*,p.version,p.filename,p.type from service s,page p where p.section_id=? and s.name=? and p.page_id=s.page_id",$sid,$modname);
		my %vserv = &keywise("version",\@service);
		#ok(@service > 0,"$modname has service page(s)");
		foreach my $v (@version) {
		    if ($v eq $config{default_version}) {
			ok($vserv{undef}||$vserv{''},"$modname has $config{default_version} service page");
		    }
		    else {
			ok($vserv{$v},"$modname has $v service page");
		    }
		}
	    }
	}
    }
}
