#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Config;
use strict;
use myConfig;
use ExSite::Store;
use ExSite::Cookie;
use ExSite::Message;

#=====================================================================
# Exported names:

BEGIN {
    use Exporter;
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

    $VERSION = "3.8";

    @ISA = qw(Exporter);

    @EXPORT      = qw(%config %share %session %store %cookie %msg %token
		      &exsite_init &exsite_close 
		      &page_header &page_footer
		      &get_obj &get_versions &preference
		      &diagnostic &OnError &AtError
		      );

    @EXPORT_OK   = qw(&read_exsite_conf 
		      &fetch_diagnostics &show_diagnostics 
		      &install_handlers);

}

use vars (@EXPORT, @EXPORT_OK);

#=====================================================================
# Non-Exported names:

use vars qw(
	    %server_config %site_config %auth_config 
	    %form_config %report_config %log_config
	    %crypt_config %prog_config
	    %search_config 
	    );

##################################################################
################ static configuration            #################
##################################################################

# static configuration parameters are valid for every site setup.
# They may still be overridden in particular scripts, by setting
# the given parameters to a new value.

#======================================================================
# SITE CONFIGURATION:
# parameters that define the website as a whole

%site_config = (

    # site name - used as a default title
    name => "ExSite",

    # site ID (no spaces) - used as a cookie prefix
    id => "ExSite",

    # identity cookie name
    identity_cookie => "ExSiteID",

    # site logo - optional <IMG ...> tag for login.cgi
    logo => "",

    # sysadm's email, eg. "webmaster\@yourdomain.com"
    administrator => undef,

);

#======================================================================
# AUTHENTICATION CONFIGURATION
# parameters controlling, login, authorization, etc.

%auth_config = (

    # authentication type (fixed, none, cookie, crypt, session)
    type => "cookie",

    # define the password storage method: 
    # cleartext - no encryption
    # des - password hashed with system crypt() function
    # md5 - password hashed with Digest::MD5
    # md5+ - multi-pass password hash using Digest::MD5
    # crypt - password encrypted using ExSite::Crypt (reversible)
    password_storage_method => "cleartext",

    # password recovery method
    # reminder - password emailed to user (not all storage methods allow this)
    # reset - password is changed and then emailed to user
    # set - user gets a link to a set password form
    # undef - use reminder if possible, and reset otherwise
    password_recover_method => undef,

    # enforce password strength on admin forms
    strong_passwords => 0,

    # system salt - can improve crypto of some things with extra salt
    salt => undef,

    # authentication scope
    # if local, users can only log into their own section
    # if global, users can log in to any section
    scope => "local",

    # define the default identity to use in case of fixed auth type, 
    # or for simple shell-based tests with no cookie handling.
    default_identity => {
	name => "nobody",
	uid => 1,
	gid => 1,
	access => 3,
    },

    # define the tables and columns where login and authentication
    # information can be found
    user_table => "member",
    group_table => "section",
    user_login_column => "login",
    user_pwd_column => "password",
    user_access_column => "access",
    user_uid_column => "member_id",
    user_gid_column => "section_id",
    user_sort_column => "name",
    user_name_column => "name",
    user_email_column => "email",

    # what authorization level corresponds to unlimited access?
    root_auth_level => 4,

    # what user level is permitted to link to records they don't own?
    foreign_link_level => 1,

    # login targets 
    # WARNING: reconfigure these if changing $config{prog}
    login_destination => [ "/index.html", 
			   "home.cgi",
			   "home.cgi",
			   "home.cgi",
			   "home.cgi",
			   ],

    # login form
    login_heading => "Login",
    login_prompt => "Login",
    password_prompt => "Password",
    input_size => 20,
    insecure_login_warning => 1,

    # include a "forgot login?" function?
    forgot_login => 1,
    forgot_prompt => "Forgot your login/password?",
    #forgot_link =>    # hard URL for forgot password link

    # include a "Remember Me" feature?
    remember_me => 1,
    remember_prompt => "Remember me on this computer",

    # how long do durable cookies last?
    long_cookie_duration => "2 weeks",

    # warn user on logout?  (leave off if you use IE6)
    logout_warning => 0,

    # clearing session
    logout_kills_session => 1,
    login_kills_session => 0,

    # check http referer on database peeks (eg. getdata.cgi, content.cgi)
    referer_check => 1,
    # ditto, but for logged-in-users (set to 0 to skip referer check for 
    # members - useful to exclude robots only from database peeks)
    member_referer_check => 1,
    # allow admins to bypass check
    owner_referer_check => 0,

    # password_access:
    # set to the access level that should be able to see password values
    # in reports and forms; note that password values are not necessarily 
    # the same as the actual password (eg. not with hashed passwords)
    password_access => 4,

    # TRUSTED CONNECTIONS
    # Do not trust connections by users of [trust_level] or higher
    # unless they are coming from an IP that matches pattern in 
    # [trusted_hosts]. Untrusted connections are downgraded to one 
    # less than the trust level. Note that trust is tested when entering 
    # admin screens (eg. home.cgi, ctrl-panel.cgi), but not necessarily
    # when accessing normal pages. (Thus authkeys should continue to 
    # work from any location when using normal pages.)
    trust_level => 4,
    trusted_hosts => '^192\.168\.',

    # email templates for password recovery
    email => {
	nouser => "Sorry, there is no user named [[login]].",	 
	twousers => "ERROR: there is more than one user named [[login]].",	 
	pleasecall => "Please contact the site administrator at [[admin]] for assistance.",	 
	missingemail => "Sorry, there is no email address recorded for the user [[login]].",	 
	email_subject => "Login instructions for [[site_title]]",
	# following are used to build email bodies
	email_body => "<p>Here are login instructions for [[user]] for the website [[site_title]]:</p>",
	email_instructions => "<p>Login: [[login]]<br>\nPassword: [[password]]</p>",	 
	noaccess => "<p>You have an old account, but its access has been disabled. To renable this account, contact your site administrator, at [[admin]].</p>",
	reset_body => "<p>A password reset was requested for [[user]] on the website [[site_title]].\n\nIf you did not make this request, you can ignore this email and your password will not be affected.\n\nTo reset your password, click on the following link:</p>",	 
	reset_instructions => "<p><a href='[[reset_url]]'>Reset password for login/user [[user]].</a></p>",
	reset_sent => "<p>Your password has been reset, and the new password has been emailed to [[email]].  You should change this password after logging in.</p>",	 
	email_error => "<p>We found an account, but your login information (username, password, or access level) is incomplete.  Please contact your site administrator for assistance, at [[admin]]</p>",	 
	notmyemail => "<p>Instructions have been emailed to [[email]]. If this is not your current e-mail address, then please contact the site administrator at [[admin]] for assistance.</p>",
	novaliduser => "<p>No valid user account was found. Please contact the site administrator at [[admin]] for assistance.</p>",
    },
);

#======================================================================
# FORM CONFIGURATION
# parameters controlling form layout

%form_config = (

    # input field separator character
    sepchar => '#',

    # multiple-input separator character
    multi_sep => '; ',

    # maximum input field width
    max_cols => 60,
    # maximum rows in a textarea input
    max_rows => 10,
    # other field width
    max_other => 10,

    # stack set checkboxes with more than this number of values
    stack_checkboxes => 4,
    stack_maxchar_per_button => 20,
    
    # use select inputs for set datatypes in search mode
    switch_search_inputs => 1,
    
    # label for select input null value
    select_nullvalue => "== select ==",

    # label for foreign key input null value    
    select_foreign_nullvalue => "== select [[title]] ==",

    # form mode:
    # "cgi" to have forms processed by CGI programs (eg. doform.cgi)
    # "exsite" to have forms processed by WebDB DCD
    mode => "cgi",

    #---------------------------------
    # dates & times
    #---------------------------------

    # default date settings for undefined dates: "today" or "null"
    default_date => "today",

    # if using a date picker, the earliest pickable date
    date_picker_start_date => "2000-01-01",

    # date input format, "simple", "picker", or "select"
    date_input_format => "select",
    # time input format, "simple", "select", or "selecthm"
    time_input_format => "selecthm",

    # date selectors ("select" or "text")
    day => "text",
    year => "text",

    # year range in year selectors (+/- this value)
    year_range => 10,
    #year_uprange => 5,
    #year_downrange => 50,

    # hour selectors can be based on a 12 or 24 hour clock
    clocktype => 12,

    # minute step size in minute selectors
    minstep => 15,

    # validation range for years (year values must be within this number
    # of years from the present)
    validate_year_range => 100,

    # can timestamp fields be edited?
    editable_timestamps => 0,

    # AM/PM labels
    am => "AM",
    pm => "PM",

    # optional datepicker localization
    datepicker_language => { 'Français' => "/_ExSite/js/date_fr.js", },

    #---------------------------------
    # foreign key inputs
    #---------------------------------

    # display of foreign key selectors
    # set to 1 to include foreign key id in selector
    # set to 0 to use only general description of foreign record
    show_foreign_key_id => 1,

    # maximum number of foreign keys in a select tag before switching
    # to a regular text input.  Set to 0 for no maximum.
    max_foreign_key_select => 500,

    #---------------------------------
    # data handling options
    #---------------------------------

    # trim leading/trailing whitespace from form data before 
    # inserting/updating?

    trim_whitespace => 0,

    # which HTML editor are we using [tinymce or htmleditor]

    editor => "tinymce",

    # default settings for htmleditor

    htmleditor => {
    		cms => 0, # use text editor in cms mode?
		# cut, copy and paste commands:
		# Gecko browser: required to add additional configuration parameters to 
		# 	the file(user.js or prefs.js) in the profile directory.
		#
		# user_pref("capability.policy.default.Clipboard.cutcopy","allAccess");
		# user_pref("capability.policy.default.Clipboard.paste", "allAccess");
		#
		# All icons are turned on by default
		op_selection=>0, # cut, copy and paste
		op_change=>1, # undo and redo
		op_table=>1, # table wizard and table editor
		op_image=>1, # image dialogue
		op_image_upload => 1, # enable file upload form in the insert image/document popup
		op_link=>1, # link dialogue (1: w/ cms options; 2: external link only)
		op_dyncontent=>1, # dynamic content dialoge
		op_style=>1, # bold,italic and underline
		op_color=>1, # text color and background color
		op_alignment=>1, # justify, align left, align right and align center
		op_list=>1, # ordered list and bulleted list
		op_indent=>1, # outdent, indent
		op_cleanup => 1, # html scrubber
		op_accent=>1, # accent drop-down list
		op_symbol=>1, # symbol drop-down list
		op_math=>1, # math drop-down list
		complete_html_content=>0,
		editor_width=>700,
		editor_height=>512,
    },

    tinymce => {
		# Tinymce path
		path=>"/_ExSite/js/tiny_mce",
		file=>"tinymce.min.js",
		# Dialog settings
		dialog=>{
			# Access level requirement
			access=>2,
			dlgimg=>{
				script=>"dlg.cgi/Image",
				allow_file_upload=>1,
				# Popup window dimension
				popup=>{
					full=>{width=>520,height=>450},
					basic=>{width=>520,height=>350},
				},
			},
			dlglink=>{
				script=>"dlg.cgi/Link",
				# Popup window dimension
				popup=>{
					full=>{width=>480,height=>300},
					basic=>{width=>480,height=>180},
				},
			},
			dlgdyn=>{
				script=>"dlg.cgi/DynContent",
			},
		},
		# Editor modes
		mode=>{
			# Current mode
			current=>"standard",
			default=>"standard",
			# Define tool groups to show in the tool bar
			toolbar=>{
				basic=>{
					toolbar_row1=>["undo","textstyle","list","link","image","remove_format","header"],
					toolbar_row2=>[],
					toolbar_row3=>[],
					toolbar_row4=>[],
					menubar=>0,
					menu=>[]
				},
				standard=>{
					toolbar_row1=>["undo","textstyle","list","link","image","remove_format","header"],
					toolbar_row2=>[],
					toolbar_row3=>[],
					toolbar_row4=>[],
					menubar=>"edit insert format view table list",
					menu=>{
						edit=>{
							title=>"Edit",
							items=>"undo redo | cut copy paste pastetext | selectall | searchreplace | code"
						},
						format=>{
							title=>"Formats",
							items=>"e_header e_block e_inline e_align | e_indent e_outdent | e_fontselect e_fontsizeselect | e_forecolor e_backcolor | e_special | removeformat"
						},
						insert=>{
							title=>"Insert",
							items=>"e_dlgimg e_dlglink e_dlgdyn media e_widget | e_dlglink_unlink | charmap hr anchor pagebreak insertdatetime nonbreaking"
						},
						table=>{
							title=>"Table",
							items=>"inserttable tableprops deltetable | cell row column | grid_list grid_add"
						},
						list=>{
							title=>"List",
							items=>"e_numlist e_bullist"
						}
					},
				},
			},
		},
		# Tool groups
		group=>{
			textstyle=>"bold,italic",
			alignment=>"alignleft,aligncenter,alignright,alignjustify",
			clipboard=>"pasteword",
			table=>"tablecontrols",
			list=>"bullist,numlist",
			format=>"block_format,inline_format",
			indent=>"outdent,indent",
			undo=>"undo,redo",
			link=>"e_dlglink,e_dlglink_unlink,anchor",
			html=>"removeformat,code,snippet",
			font=>"fontselect,fontsizeselect",
			color=>"forecolor,backcolor",
			exsite=>"e_dlgdyn",
			special=>"hr,charmap,nonbreaking",
			layer=>"insertlayer,moveforward,movebackward,absolute",
			attribute=>"styleprops",
			direction=>"ltr,rtl",
			datetime=>"insertdate,inserttime", 
			image=>"e_dlgimg", 
			expert=>"advhr,attribs,media,cleanup,pastetext",
			header=>"header_format",
			mailto=>"e_dlgmailto",
			remove_format=>"removeformat"
		},
		# Tinymce settings
		editor=>{
			width=>"600",
			height=>"300",
			theme=>"modern",
			plugins=>"autolink,lists,charmap,hr,anchor,pagebreak,visualblocks,visualchars,code,media,nonbreaking,table,directionality,emoticons,paste,textcolor,format,searchreplace,dlglink,dlgmailto",
			element_format=>"html",
			convert_fonts_to_spans=>1,
			fontsize_formats=>"8pt 10pt 12pt 14pt 18pt 24pt 36pt",
			font_formats=>"Serif=serif;Sans-serif=sans-serif;Monospace=monospace",
			entity_encoding=>"raw",
			force_p_newlines=>0,
			forced_root_block=>"p",
			extended_valid_elements=>"iframe[src|width|height|name|align|frameborder|title|class|type|allowfullscreen]",
			convert_urls=>0,
			exsite_style_formats=>"Button=button;Info=info;Highlight=highlight",
			browser_spellcheck=>1,
		}
    },

    # file upload limits (in btyes)
    # "file" type is capped by field size (64K, which is reasonable)
    # "bigfile" type allows up to 16MB, which is probably unreasonable
    # for web uploads.  Furthermore DB engines like MySQL limit query
    # sizes so that you can't actually insert large files into a SQL
    # record.  Set the actual bigfile size limit here:

    bigfile_maxsize => 4194304,  # 4 MB

    # Add extra span tags around form labels and inputs for CSS purposes.
    # If this is disabled, css classes are applied directly to label and
    # input tags.

    css_span => 1,

);

#======================================================================
# REPORT CONFIGURATION
# parameters controlling report formatting

%report_config = (

    # currency character
    currency_char => '$',

    # maximum number of records to report at a time
    page_length => 20,

    # for table reports, this flag determines whether each row is classified 
    # as A or B for the purposes of highlighting alternate lines.  Note that 
    # turning this flag on does not by itself highlight lines;  you must also 
    # setup your stylesheet to recognize the A/B classes of <TR>.  This 
    # setting will override $reportrecordtag->{table}->[0].
    highlight_alternate_lines => 0,

    # report link mode:
    # "cgi" to have db_ops linked to CGI programs (eg. webdb.cgi)
    # "exsite" to have db_ops linked back to the page (ie. to WebDB DCD)
    linkmode => "cgi",

    # foreign key linking permissions
    # when reporting foreign keys, cross-link to the foreign record if the
    # viewer has this permission on the foreign record
    link_to_foreign_key_perm => "read",

    # if primary display columns of a table include foreign keys, this 
    # setting will expand those foreign keys to readable form when 
    # building record labels
    show_foreign_keys_in_record_labels => 0,

    # if there is more than one primary display column for a table, but only
    # one is requested, this setting selects either the first or last
    use_first_key => 0,

    # email format:
    # "plain" displays the email using a regular mailto: URI
    # "obfuscated" uses javascript to obfuscate the address from spambots
    email => "plain",
    
    # display permission errors
    show_no_permission => 1,

    # dynamic reports - use jquery plugin for more interactive reports
    dynamic => 0,
    # enable/disable pagination for dynamic report
    dynamic_paginate => 1,
    # default page length (0 for default)
    dynamic_page_length => 0,

    # sorting
    case_insensitive_sort => 0,

    #---------------------------------
    # reporting of text data
    #---------------------------------

    text => {
	
	# allow HTML tags found in normal text

	honor_html => 0,

	# if HTML is allowed, allow only non-structural tags, to prevent
	# layouts from being damaged (ie.no tables or divs)

	safe_html => 1,

	# treat newlines as meaningful in normal text

	honor_nl => 0,

    },

    # date/time reporting

    date => {
	date => "date",
	time => "time",
	datetime => "datetime2",
	timestamp => "datetime2",
        localize => 0,   # put format strings in system messages, if true
    },

);

#======================================================================
# SERVER CONFIGURATION
# this stuff varies between sites, so the following defaults are
# mostly just placeholders for configuration data to follow.
# The auto flag determines whether autoconfiguration is done by
# default.  Set it to 1 in exsite.conf if you want autoconfiguration.

%server_config = (
    db => { 
	name=>"exsite", 
	server=>"localhost", 
	user=>"exsite", 
	type=>"sql", 
	charset=>"",  # set to UTF8 if your DB uses UTF-8 encoding
	map=>"map",
	engine=>"MyISAM"
	},
    map => { name=>"dbmap", type=>"text" },
    HTMLroot => "/var/www/html",
    HTMLpath => "",
    CGIroot => "/var/www/cgi-bin",
    CGIpath => "/cgi-bin",
    protocol => "http",
    host => "localhost",
    domain => "localdomain",
    port => "",
    server => "http://localhost.localdomain",
    rootserver => undef,
    temppath => "/tmp",
    imgdir => "/images",
    securepath => "/usr/local/bin:/bin:/usr/bin:/usr/X11R6/bin",
    auto => 0,
    HTMLroot_lock => 1, # prohibits writing to directories outside HTMLroot
);

#======================================================================
# LOG CONFIGURATION
# parameters controlling event logging

%log_config = (

    # Each parameter refers to the logging level for a particular log file.
    # Log levels go from 0 (no logging) to 3 (log everything).
    # Log files must exist and be writable by the webserver.

    # db - database log file
    # 1: log database errors
    # 2: log database modifications
    # 3: log database connections & selects
	       
    db => 0,

    # cms - content management system audit trail
    # 1: log publicly-visible changes
    # 2: log all website, page, and content updates
    # 3: log all cms operations

    cms => 0,

    # general - general ExSite diagnostics
    # 1: log errors
    # 2: log warnings
    # 3: log info messages

    general => 0,

    # audit - things that change system state
    # log levels are ignored on audit, any non-zero value enables it

    audit => 0,
    
    # cron
    # log levels are ignored on cron, any non-zero value enables it

    cron => 0,

    # the log files themselves... (paths are relative to cgi directory)

    logfile => {
	db => "db.log",
	cms => "cms.log",
	general => "general.log",
	audit => "audit.log",
	cron => "cron.log",
	post => "post.log",	
    },

);

#======================================================================
# CRYPTO CONFIGURATION 

%crypt_config = (

    # CRYPTOGRAPHY:
    # ExSite can automatically encrypt and decrypt data going to and from
    # the database, if the column has "crypt" datatype.  Crypt fields are
    # treated same as text fields, but the data is encrypted according
    # to the following options:

    # encryption mode can be:
    # "strong" - best security, good for storing critical data
    # "fast" - trades security for speed; 2-3 times faster than strong
    #          mode; good for obfuscating data communicated over http, 
    #          but not for secure storage.

    default_mode => "strong",

    # keys for coding/decoding
    # It is good practice to define different keys for fast and strong
    # encryption.  Otherwise the regular key will be used.

    key_fast => "fast dummy pass-phrase",
    key_strong => "strong dummy pass-phrase",
    key => "dummy pass-phrase",

    # number of mixing rounds (higher is better)

    nround_fast => 4,
    nround_strong => 32,

    # decryption rules:
    # admin => only administrator can view plaintext
    # owner => owner of record can view plaintext

    access => "admin",

);

##======================================================================
# PROGRAM CONFIGURATION 

%prog_config = (
    captcha          => "captcha.cgi",
    content          => "content.cgi",
    ctrlpanel        => "ctrl-panel.cgi",
    dcd              => "dcd.cgi",
    debug            => "debug.cgi",
    dlg_dyncontent   => "dlg_dyncontent.cgi",
    dlg_image        => "dlg_image.cgi",
    dlg_link         => "dlg_link.cgi",
    dlg_table        => "dlg_table.cgi",
    dlg_table_editor => "dlg_table_editor.cgi",
    doform           => "doform.cgi",
    echo             => "echo.cgi",
    getdata          => "getdata.cgi",
    help             => "help.cgi",
    home             => "home.cgi",
    login            => "login.cgi",
    logout           => "logout.cgi",
    page             => "page.cgi",
    publish          => "publish.cgi",
    viewcontent      => "viewcontent.cgi",
    view             => "view.cgi",
    webdb            => "webdb.cgi",
);

#======================================================================
# SEARCH CONFIGURATION 

%search_config = (

    # scope = local (search current section only) or global (search all)

    scope => "local",

    # do not index the words in the following file

    skipwords => "skipwords.txt",

    # alternate method of specifying skipwords
#    skipwords => ['a', 'an', 'the', 'to', 'of', 'it', 'as', 'and', 'or', 
#		  'is', 'are', 'am', 'be', 'on', 'in', 'at', 'if', 'by', 
#		  'do', 'for', 's', 'so', 't', 'this', 'that', 'these', 
#		  'those', 'they', 'he', 'she', 'we', 'i', 'were', 'what', 
#		  'with', 'nt', 'you', 'was', 'his', 'her', 'have', 'from',
#		  'had', 'but', 'not', 'all', 'we', 'when', 'your', 'can',
#		  'said', 'there', 'use', 'each', 'which', 'how', 'their'],

    # should we index numbers?

    index_numbers => 1,

    # when indexing numbers, ignore those below a certain value;
    # this ignores all 1- and 2-digit numbers

    min_number => 100,

    # weight words in different parts of the page

    weight => {
	title => 10,
	filename => 5,
	keywords => 5,
	description => 3,
	body => 1,
    },

    # default maximum number of results to report

    maxresult => 25,

    # threshold (minimum weight) of indexed terms

    threshold => 1,

    # automatically reindex pages and sites when they are published

    reindex_on_publish => 0,

    # no results message

    no_results => "Nothing relevant was found.",

    # do we index archived pages?

    index_archived => 0,
    
    # html placeholder text for search input
    
    placeholder_text => "Search",

);

#======================================================================
# GENERAL CONFIGURATION 

%config = (

    #---------------------------------
    # HTML output options:
    #---------------------------------

    # doctype of system screens - default to loose because who knows
    # what plug-ins will try to throw into our screens??  But if you
    # have a little control over your setup and which plugins you are
    # using, feel free to get stricter.  If you use XHTML, be sure to
    # set the markup rules (below) to enable XML syntax.

    doctype => '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">',

    # The character set used for admin screens and the editor.  UTF-8 allows
    # for native representation of most languages; ISO-8859-1 is a common
    # alternative, but is only good for Western European Languages.

    charset => "UTF-8",

    # webtop help link

    help_url => "http://support.exsitewebware.com",

    # webtop icon set (blank for default/classic)

    iconset => '',

    # standard jquery library

    jquery=>"//ajax.googleapis.com/ajax/libs/jquery/1.8/jquery.min.js",
    jqueryui=>"//ajax.googleapis.com/ajax/libs/jqueryui/1.8/jquery-ui.min.js",

    # The default language/version.

    default_version => "English",

    # version suffixes (defaults to 1st 2 chars of version name,
    # but may need to be defined in cases of Asian charsets where
    # this might not result in clean characters)
    # eg. version_extension.中文 = Ch

    version_extension => {},

    # highlight any missing message translations

    highlight_missing_sysmsg => 1,

    # PAGE BUFFERING
    # set true if using the page buffering module, ExSite::PageBuf.pm, 
    # by default;  set false if output goes straight to stdout by default.
    # NB: page buffering can be changed from the default for specific scripts 
    # by setting this flag at the top of the script.

    page_buffering => 0,

    # Default thumbnail size (max edge length in pixels)

    thumbnail_size => 100,

    # Maximum image size (max edge length in pixels)

    max_image_size => 1600,

    # DATABASE OPTIONS
    # handling of database images
    # set true if ExSite::Report::show_data should display images inline
    # set false if ExSite::Report::show_data should display a link to the image

    show_images => 1,

    # Timestamp format;  values are:
    # "num" (eg. 20070411161709, MySQL 3)
    # "text" (eg. 2007-04-11 16:17:09, MySQL 4.1)

    timestamp_format => "text",

    # DBmap type

    dbmap_type => "external",

    # maximum upload file size
    # NB: form.bigfile_maxsize limits the size of a file that can be loaded
    # into a bigfile DB record.  This parameter, on the other hand, limits 
    # the size of a file that can be uploaded through a web form for other
    # reasons (eg. publish straight to disk).

    max_upload_size => 16000000,

    # maximum tree size (in nodes)
    # This is mostly to prevent node loops in trees.

    max_tree_size => 10000,

    # mail notifications
    # method can be "sendmail" or "Net::SMTP" (preferred)
    # sendmailpath only used for "sendmail" method
    # bcc can be used to send copies of all notifications to an archive PO box
    # set bulklevel and bulkserver to route mail through alternate servers
    # if the volumn of messages is high

    mail => {
	method => "Net::SMTP",
        sendmailpath => "/usr/lib/sendmail",
	bcc => undef,
	server => "localhost",
	bulkserver => "localhost",
	bulklevel => 2,  # max number of messages to send thru regular server
        dummy_html_part => 0,  # create an HTML part for plain text emails sent in multipart mode
        firewall => undef,  # divert all emails to this address, if defined 
    },

    # section handling by CMS

    section => {
        standalone_subdir_404 => 0,       # standalone subdirs give 404
    },

    # page handling by CMS

    page => {
	suffix => '\.(html?|php|asp)',    # allowed filename suffixes
	dynamic_method => "id",           # id, path
        dyn_publish_method => "none",     # none, redirect, template
	editmode => "MySite",             # editable page prev. handler
	static_urls => 0,                 # always use static urls?
	static_url_type => "full",        # use full static urls
	dynamic_url_type => "relative",   # use relative dynamic urls
        default_to_index => 1,            # go to index if dyn path not found
        no_index_filename => 0,           # don't specify index.html in urls
        unused_path_error => "warn",      # how to handle unknown residual PATH_INFO
        no_page_404 => 0,                 # enable to 404 if no page to display
# these are superceded by dyn_publish_method = template|redirect
#	prepublish_dynamic => 0,          # 
#	redirect_dynamic => 0,            # publish dynamic page as a meta
#                                          # page that refreshes to dynamic
# these are superceded by static_url, dynamic_url settings
#	relative_static_url => 0,         # use relative static urls
#	relative_dynamic_url => 1,        # use relative dynamic urls
    },

    # content handling by CMS

    content => {
	inline_css => 0,                         # place CSS in <style> block
	preformatted_plaintext => 1,             # place plaintext in <pre> blk
	img_title => undef,                      # image title can be name,
	                                         #   description, or undef
        block_unpublished => 0,                  # block viewing of unpublished content
    },

    # captchas

    captcha => {
	mode => "image",                         # default captcha type

	# image captchas
	font => "ps:Courier-Bold",               # font to use
	pointsize => 30,                   	 # size of text
	color => "#333366",                      # color of text
	distort => ".25",                        # distortion factor
	start_x => 5,                            # start position of text
	start_y => 30,                           # "
	charsize => 19,                          # character width

	# passive captchas
	passive_method => "honeypot",            # postkey,honeypot,js
	postkey_cookie => "postkey",             # name of postkey cookie
	honeypot_name => "comment",              # honeypot field
	max_postkey => 3,
        js_name => "_pc",                        # javascript form field
        js_value => 1,                           # javascript form value

	# text captchas...
	dictionary => "/usr/share/dict/words",   # source of words
	word_set_size => 5,                      # number of words in a challenge
	max_password_size => 2,                  # number of characters required in a response

	# time-sensitive
	expire => 24,                            # set to >0 to prevent re-use
	                                         # after that many hours
    },

    # error handling

    diagnostics => {
        level => { info=>1, warning=>2, error=>3, critical=>4, fatal=>5 },
        severity => [ undef, "info", "warning", "error", "critical", "fatal" ],
        # display the severity level as a prefix when displaying diagnostics
        show_severity => 1,    
    },

    # content filters

    wordfilter => {
	mode => "off",                           # off, auto, custom
	method => "reject",                      # reject, censor, semi-censor
	file => "badwords.txt",                  # file in conf
	censor_string => "[censored]",           # used for action=censor
	censor_char => "*",                      # used for action=semi-censor
	threshold => 1,                          # read level to act on
	override => 3,                           # access level to ignore
    },

    # handling of database operations
    # each operation has the following parameters:
    #   approve: the low-level operation that needs to be approved by DB::approve
    #   url: the script that executes the operation
    #   img: the image for the operator icon
    #   img_width, img_height: the dimensions of the image
    #   label: the text string to use if the image is not present
    #   table: 1 means operation can be performed on whole tables
    #          (otherwise, op is restricted to individual records)
    # Each url is passed ?tab=<table>[&id=<record>]
    #
    # WARNING: reconfigure these if changing $config{prog}

    db_ops => {
	edit => { approve=>"update", url=>"?action=edit", 
		  img=>"edit12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"edit", table=>0 },
	copy => { approve=>"insert", url=>"?action=copy", 
		  img=>"copy12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"copy", table=>0 },
	delete => { approve=>"delete", url=>"?action=delete", 
		  img=>"rm12.gif", level=>3,
	          img_width=>12, img_height=>13, label=>"delete", table=>0 },
	new => { approve=>"insert", url=>"?action=new", 
		  img=>"new12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"add new", table=>1 },
	view => { approve=>"read", url=>"?action=view", 
		  img=>"view12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"view details", table=>0 },
	list => { approve=>"read", url=>"?action=list", 
		  img=>"list12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"list", table=>1 },
	report => { approve=>"read", url=>"?action=report", 
		  img=>"list12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"list", table=>1 },
	searchquery => { approve=>"read", url=>"?action=query", 
		  img=>"search12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"search", table=>1 },
    },

    db_ops_grps => {
	#all => [ "list", "view", "search", "edit", "copy", "new", "delete" ],
	all => [ "list", "view", "searchquery", "edit", "copy", "new", "delete" ],
	all_edit => ["edit", "copy", "delete"],
	all_view => ["list", "view", "search"],
	link => ["view"],  # kludge to support old "link" args
    },

    # preferred order for displaying db operators, when more than one is shown
    # ops not in this list will not be displayed at all.

    #db_ops_order => [ "list", "view", "search", "edit", "copy", "new", "delete" ],
    db_ops_order => [ "list", "view", "searchquery", "edit", "copy", "new", "delete" ],

    # HTML snippets (user-defined) [DEPRECATED]

    snippet => {},

    # include the other configuration hashes here

    site => \%site_config,
    auth => \%auth_config,
    form => \%form_config,
    report => \%report_config,
    log => \%log_config,
    crypt => \%crypt_config,
    prog => \%prog_config,
    search => \%search_config,
    server => \%server_config,  # conditional

    # CACHE

    cache => {
	max_size => 1000000,    # only cache items under 1 MB
	persistent => 1,        # try to save cache beyond the current request
#	querytype => {
#	    get_columns => 1,
#	    get_key => 1,
#	    count => 1,
#	    fetch => 1,
#	    fetch_match => 1,
#	    fetch_all => 1,
#	    fetch_child => 1,
#	    fetch_m2m => 1,
#	    select => 1,
#	},
    },

    # SESSIONS

    session => {
	max_idle => 3600,       # purge sessions after this much idle time
        secure => 0,            # enable anti-hijacking features
    },

    # markup rules

    markup => {
	html => 1,              # apply some extra HTML formatting rules
	xml => 0,               # use XML syntax
	minattr => 1,           # minimize attributes, if possible
	safe_content => 1,      # assume contents are HTML-safe
	safe_attributes => 1,   # assume attributes are HTML-safe
	nl => 1,                # append a NL char to end of element
    },

    # CRON

    cron => {
	mday => 1,              # which day to run monthly tasks    
	day => 0,               # which day to run weekly tasks
	hour => 1,              # which hour to run daily tasks
    },

    # TOKENS

    token => {
	type => "store",        # store or db
	captcha => 0,           # use for captchas
	short_url => 1,         # use for short URLs,
	authtoken => 0,         # use for authtokens
	secure_query => 0,      # use for secure query strings
	secure_query_duration => 2592000, # 30 days
    },

    # Timezones

    timezone => {
        auto => 0,             # autoconf server timezone to detect DST
        server => "-0800",      # default server timezone
        home => undef,          # offset of website owner from server timezone
    },

    #---------------------------------
    # kill switches
    #---------------------------------

    # kill - prevents page.cgi from generating pages
    # killurl - redirect killed page.cgi requests to this URL

    kill => 0,
    killurl => undef,

    # busy - prevents page.cgi from generating pages if load is too high
    # busyurl - when busy, redirect requests to this URL
    # busyload - the 1-minute load average threshold that triggers redirects
    # NOTE: we check /proc/loadavg for the load value; this feature only
    # works on systems that have this file.

    busy => 0,
    busyurl => undef,
    busyload => 10,

    # throttle - delay requests if rate is excessive
    # rate - rate (req/s) at which to throttle requests (0 to disable)
    # threshold - interval (s) at which recent requests are considered potentially abusive
    # min - minimum number of abusive requests to trigger throttle
    # url - redirect to this URL when being throttled
    # history - number of recent requests to track
    #
    # WARNING: if you set the rate to less than 0.1, the 10 s automatic
    # refresh of the default throttle page will keep them in perpetual 
    # refresh hell.  0.2 req/s is okay for users, 2 req/s is okay for sites.
    throttle => {
	user => {
	    rate => 0,
	    threshold => 10,
	    min => 4,
	    url => undef,
	    history => 10,
	},
	system => {
	    rate => 0,
	    threshold => 2,
	    min => 10,
	    url => undef,
	    history => 20,
	},
    },

    # webtop
    webtop => {
        # access level required to reach webtop
        access => 2,
        # restrict level 2 admins to a whitelist of allowed modules
        whitelist => undef,
        # should we add a section selector to the webtop?  Set to:
        # "Y" to add an unset selector,
        # "home" to preset to the user's home section,
        # or a section ID to preset to that section for all users
        section_selector => 0,
        # where to send users who are denied webtop access 
        # (menu, login_destination)
        deny => "login_destination",
        # popup style:
        # popup - opens control panels in simple popup window
        # tab - opens control panels in a new tab (actually a new window, but
        #       some browsers are configured to open new windows in tabs)
        popup_style => "popup",
        # categories
        categories => [ "CMS", "Applications", "System", "Database" ],
        headings => [ "Content Management", "Applications", "System", "Database" ],
        # NOTE: the "Ecommerce" category is also supported, but not enabled
        # by default
        # allow manual renaming of modules, eg.
        # webtop.module_name.Foo = Bar
        # If a module is renamed to undef/nothing, it should not be shown.
        module_name => { BaseDCD=>undef },
        # favourites: if defined, add a Favourites category, and put the 
        # following modules into it:
        #favourites =>  "MySite,Blog",
        favourites_heading => "Favourites",
        # topdcd - if defined, place this module at the start of its row
        #topdcd => { CMS => "MySite" },
    },

    # what modules should be included in the Insert Web App selector
    dcd => {
        whitelist => undef,
        blacklist => ["AddrBook","Adjustment","Backup","BillPayment","CMS",
                      "Cron","Dashboard","Email","Finance","Giftcard",
                      "ObjectBrowser","Orders","Permissions","Preferences",
                      "POS","Query","Security","SEO","StoreAdm","SysMsg",
                      "ToDo","Trash","Upload","Webalizer","WebDB"],
    },

);

#### end of static configuration


##################################################################
################ setup and configuration routines ################
##################################################################

sub server_autoconf {
    $config{server}{HTMLroot} = $ENV{DOCUMENT_ROOT};
    $ENV{SCRIPT_NAME} =~ m:^.*/([^/]+)$:;
    my $script_name = $1;
    $ENV{SCRIPT_FILENAME} =~ /^(.*)\/$script_name$/;
    $config{server}{CGIroot} = $1;
    my $script_re = join('|',values %{$config{prog}},'[\w-]+\.cgi');
    $config{server}{CGIpath} = ($ENV{SCRIPT_NAME} =~ /^(.*)\/($script_re)$/)? $1 : "";
    $config{server}{protocol} = exists $ENV{HTTPS} ? "https" : "http";
    $config{server}{HTMLpath} = $config{server}{CGIpath};
    ### FIXME: support other cgi paths
    $config{server}{HTMLpath} =~ s/\/cgi(-bin)?\//\//;
    $config{server}{HTMLpath} =~ s/\/cgi(-bin)?$/\//;
    if ($config{server}{HTMLpath} eq "/") {
        $config{server}{HTMLpath} = "";
    }

    # new method - handles 2- or 3-part names
    $ENV{HTTP_HOST} =~ /^((.+?)\.)?(.*?\..*?)(:\d+)?$/;
    $config{server}{host} = $2;
    $config{server}{domain} = $3;
    $config{server}{port} = $4;
}

sub read_exsite_conf {
    my ($file,$key) = @_;
    if (! $file) { $file = "exsite.conf"; }
    if ($key && ! exists $config{$key}) { $config{$key} = {}; }
    my @confpath = ("./conf", "ExSite", ".");
    foreach my $dir (@confpath) {
	if (-e "$dir/$file") {
	    if (open CONF,"<$dir/$file") {
		foreach (<CONF>) {
		    next if (/^\#/);
		    next if (/^\s*$/);
		    chomp;
		    if (/^include <(.+)>/) {
			&read_exsite_conf($1,$key);
			next;
		    }
		    my ($param,$value,$append);
		    if (/\+=/) { 
			($param,$value) = split /\s*\+=\s*/,$_,2;
			$append = 1;
		    }
		    else {
			($param,$value) = split /\s*=\s*/,$_,2;
			$append = 0;
		    }
		    $value =~ s/\\n/\n/g;
		    my @param = split /\./,$param;
		    my $conf = $key ? $config{$key} : \%config;
		    foreach my $par (@param[0..$#param-1]) {
			if (! exists $conf->{$par}) {
			    $conf->{$par} = {};
			}
			# this will crash if not a hash ref
			$conf = $conf->{$par};
		    }
		    if ($append) {
			# append parameter to list
			if (! exists $conf->{$param[-1]}) {
			    $conf->{$param[-1]} = [$value];
			}
			else {
			    if (ref $conf->{$param[-1]} ne "ARRAY") {
				my $tmplist = [ $conf->{$param[-1]} ];
				$conf->{$param[-1]} = $tmplist;
			    }
			    push @{$conf->{$param[-1]}},$value;
			}
		    }
		    else {
			# simple parameter
			$conf->{$param[-1]} = $value;
		    }
		}
		close CONF;
	    }
	}
    }
}

# Preferences
#
# Preferences are localized configuration settings - specific to a
# certain context (eg. for a section, or a user).  They are stored in 
# the attributes table.
#
# If no preference is found, the system configuration setting is used.
#
# confname is a configuration setting, eg. "foo.bar"
# @context is a list of table/record ID pairs, eg. page 10, section 1.
# If no context is given, we assume the current section (based on the
# page we are looking at).  We search for an attribute named $confname, 
# attached to any of the table/record pairs in the context.

sub preference {
    my ($confname,@context) = @_;
    # confname = group.subgroup.name
    my @confname = split /\./,$confname;
    if (@context == 0) {
	# default context is current section
	my $this_section = $share{DB}->this_site();
	if ($this_section) {
	    @context = ( "section", $this_section->{section_id} );
	}
    }
    while (@context) {
	my $context_type = shift @context;
	my $context_id = shift @context;
	if ($context_type && defined $context_id) {
	    if (! $share{config}{$context_type}{$context_id}) {
		# fixme: check the store
		my @attr = $share{DB}->fetch_match("attribute",
						   {tablename=>$context_type,
						    id=>$context_id});
		foreach my $attr (@attr) {
		    $share{config}{$context_type}{$context_id}{$attr->{name}} = $attr->{value};
		}
		# fixme: recache in the store
	    }
	    if (exists $share{config}{$context_type}{$context_id}{$confname}) {
		# aleays return a defined value, even if the DB value is null
		# This is so we can tell the difference between a preference
		# that has been set to nothing, and no preference.
		my $val = $share{config}{$context_type}{$context_id}{$confname};
		return $val ? $val : "";
	    }
	}
    }
    # return system setting
    my $conf = \%config;
    while (my $name = shift @confname) {
	return undef if (ref $conf ne "HASH");
	$conf = $conf->{$name};
    }
    return $conf;
}

############################################################
################ initialization and shutown ################
############################################################

sub exsite_init {

    # reset share hash
    %share = (
	      diagnostics => [],
	      ENV => {},
	      );

    # set up the data store
    &exsite_init_store;

    # load configurations
    &exsite_init_config;

    # fill cookie jar
    &exsite_init_cookie;

    # load translations
    &exsite_init_msg;

    # initalize session
    &exsite_init_session;

    # setup tokens - optional feature for now
    if (eval 'require ExSite::Token') {
	tie %token, 'ExSite::Token';
    }

    # save our original environment state in case we need to tamper with it
    $share{ENV}{SCRIPT_NAME} = $ENV{SCRIPT_NAME};
    $share{ENV}{QUERY_STRING} = $ENV{QUERY_STRING};
    $share{ENV}{PATH_INFO} = $ENV{PATH_INFO};
    $share{ENV}{REQUEST_METHOD} = $ENV{REQUEST_METHOD};

    # timezone config
    if ($config{timezone}{auto}) {
	if (! $store{timezone}) {
	    my $tz = `date +%z`;
	    chomp $tz;
	    $tz =~ /^([\-\+]\d+)$/;
	    $store{timezone} = $1;
	}
	$config{timezone}{server} = $store{timezone};
    }

    # localized initialization
    &my_exsite_init(\%config,\%share,\%session,\%store,\%cookie,\%msg,@_);

    # untaint our publication directories
    {
	$config{server}{HTMLroot} =~ /^((\/[\w\.-]+)+)$/;
	$config{server}{HTMLroot} = $1;
    }
    {
	$config{server}{HTMLpath} =~ /^((\/[\w\.-]+)*)$/;
	$config{server}{HTMLpath} = $1;
    }
    {
	$config{page}{suffix} =~ /^(.+)$/;
	$config{page}{suffix} = $1;
    }
    # untaint our secure path
    {
	$config{server}{securepath} =~ /^(((\.|(\/\w+)+):)+(\/\w+)+)$/;
	$config{server}{securepath} = $1;
    }
}

sub exsite_init_store {
    # set up the data store
    my ($store_conf,$st);
    eval { 
	$store_conf = &Local::store_conf() 
	    or die 'no store configuration';
    };
    if (! $@ && ref $store_conf eq "HASH") {
	# looks like we have a data store; try connecting to it
	$st = tied %store;  # might still be tied in a persistent perl setup
	if (! $st) { 
	    # not tied, so re-tie it
	    $st = tie %store, 'ExSite::Store', $store_conf;
	}
    }
    else {
	# persistent store not enabled
	# Note that %store contents may persist in some environments 
	# (eg. Persistent Perl);  however we cannot rely on these contents
	# because it is local to the process, not the server (ie. if another
	# process changes the store contents, those changes are not reflected
	# here.)  Therefore we must clear the store on each request.
	%store = ();
    }
}

sub exsite_init_config {
    if (! $config{_loaded}) {
	# %config not preloaded - must reload from scratch

	# load any extra configuration from conf files
	&read_exsite_conf();

	# store configuration persistently
	$config{_loaded} = 1;
	my $st = tied %store;
	if ($st) { $st->put("config:",\%config,0); }
    }
    if ($config{test_conf}) {
	# the test_conf file allows us to override std configs; this is 
	# typically used on dev/test installs, so that we don't have to
	# merge/sync conf files with the production site
	&read_exsite_conf($config{test_conf});
    }

    # run autoconfiguration, in case of flexible URLs/domains
    if ($config{server}{auto}) {
	&server_autoconf;
    }

    # build the server address from the various config params
    if ($config{server}{host}) {
	$config{server}{server} = "$config{server}{protocol}://$config{server}{host}.$config{server}{domain}$config{server}{port}";
    }
    else {
	$config{server}{server} = "$config{server}{protocol}://$config{server}{domain}$config{server}{port}";
    }

    # patch up our db_ops urls, in case we've rewritten our script names
    foreach my $op (keys %{$config{db_ops}}) {
	$config{db_ops}{$op}{url} = $config{prog}{webdb}.$config{db_ops}{$op}{url};
    }
}

sub exsite_init_cookie {
    # set up the cookie hash now that we have some server info
    if (! $config{site}{cookiepath}) {
	# default to send cookies only to CGI programs
	# set site.cookiepath to "/" if you need cookies on your
	# static pages, or you are doing URL rewrites
	$config{site}{cookiepath} = $config{server}{CGIpath};
    }
    my $domain = $config{server}{host} ? 
	".$config{server}{domain}" : 
	$config{server}{domain}; 

    if ($config{strip_cookie_end_semicolon}) {
	# workaround for defective cookie implementations
	$ENV{'HTTP_COOKIE'} =~ s/;$//;
    }

    tie %cookie, 'ExSite::Cookie',
        $config{site}{cookiepath},
        $domain,
        $config{auth}{long_cookie_duration};
}

sub exsite_init_msg {
    # setup system messages
    tie %msg, 'ExSite::Message', $config{default_version}, $config{highlight_missing_sysmsg}, \%share, \%store;
}

sub exsite_init_session {
    my $st = tied %store;
    if ($st) {
	# session management is automatically enabled if there is 
	# a persistent data store
	require ExSite::Session;
	tie %session, 'ExSite::Session';
    }
}

sub exsite_close {
    # localized clean-up;
    &my_exsite_close(@_);

    # flush the session
    if (tied %session) { untie %session; }

    # flush the cache
    if ((ref $share{Cache}) =~ /Cache/) { $share{Cache}->flush; }

    # shut down the store
    if (tied %store) { untie %store; }

    # close primary database
    if ($share{DB}) { $share{DB}->close(); }

    # wait for any child processes to wrap up
    do {} until (wait == -1);
}

sub reset_share {
    foreach my $item (keys %share) {
	# keep items whose state does not change
	next if ($item =~ /^(DB|Cache|content|config|ENV|identity|diagnostics|publish|querylib)$/);
	next if ($item =~ /^dbmap:/);
	delete $share{$item}
    }
}

##################################################################
################ object management                ################
##################################################################

sub get_obj {
    my ($type,$id) = @_;

    if ($id) {

	# instantiate an object of this type and ID

	my $obj;
	if ($share{DB}) {
	    my $class = $share{DB}{map}->get_table_attr($type,"class");
	    if ($class) {
		# object has a specialized class
		$class =~ /^((ExSite|Modules)::.*)$/;
		my $safe_class = $1;
		if (eval "require $safe_class") {
		    $id =~ /^(\d*)$/;
		    my $safe_id = $1;
		    if ($safe_id) {
			$obj = eval "new $safe_class(id=>$safe_id)";
			return $obj if ($obj);
		    }
		    &diagnostic({message=>"$class: $@",type=>"general",level=>4});
		}
		else {
		    &diagnostic({message=>"$class: $@",type=>"general",level=>4});
		}
	    }
	}
	if (! $obj) {
	    # setup as a generic object
	    return new ExSite::Object(type=>$type,id=>$id);
	}
    }
    else {

	# no id - looking for a predefinded object in $share

	if ($share{$type}) {
	    if ((ref $share{$type}) =~ /$type/) {
		return $share{$type};
	    }
	    # error: no such object, and the %share spot is occupied
	    return undef;
	}
    
	# no object found in %share;  look for class in ExSite
	if (eval "require ExSite::$type") {
	    my $o = eval "new ExSite::$type";
	    if ($o) {
		$share{$type} = $o;
		return $o;
	    }
	}

	# hm, nothing in ExSite;  look for class in Modules
	if (eval "require Modules::$type") {
	    my $o = eval "new Modules::$type";
	    if ($o) {
		$share{$type} = $o;
		return $o;
	    }
	}
    }
    
    # nothing found
    return undef;
}

##################################################################
################ version management               ################
##################################################################

sub get_versions {
    my $all = shift;  # includes default version, if true;
    my @version;
    if ($all) {
	push @version, $config{default_version};
    }
    if ($share{DB}) {
	push @version, split /\|/, $share{DB}{map}->regexp("list:version");
    }
    return @version;
}

##################################################################
################ cron                             ################
##################################################################

sub is_cron_running {
    if ($share{DB}) {
	#$share{DB}->load_submaps();  # won't work if Cron is not loaded
	if ($share{DB}{map}->is_mapped("crontask")) {
	    my @cron = $share{DB}->fetch_all("crontask");
	    foreach my $cron (@cron) {
		# return true if there are any tasks that have been run
		return 1 if ($cron->{exectime} && $cron->{exectime} !~ /^0/);
		# Technically, this will return false until the first pass through cron.
		# Also, it will return true if cron ran in the past, but has since been 
		# disabled.
	    }
	}
    }
    return undef;
}

##################################################################
################ diagnostics                      ################
##################################################################

# diagnostics have 5 lavels: informational, warning, error, critical, fatal
# level 3 (error) is presumed by default

sub diagnostic {
    my @log;
    while (@_) {
	my $message = pop;
	my $msg;
	if (ref $message) {
	    $msg = $message;
	}
	else {
	    $msg = {
		message => $message,
		type => "general",
		level => 3,
	    };
	}
	$msg->{type}  or $msg->{type} = "general";
	$msg->{level} or $msg->{level} = 3;
	my $severity = $config{diagnostics}{severity}[$msg->{level}];
	$msg->{message} = uc($severity) . ": " . $msg->{message};
	# logging
	if ($msg->{level} >= 4 - $config{log}{$msg->{type}}) {
	    # log in %share and to disk
	    push @{$share{diagnostics}}, $msg;
	    push @log, $msg;
	}
	elsif ($msg->{level} >= 3) {
	    # always log level 3 diagnostics for internal error trapping
	    push @{$share{diagnostics}}, $msg;
	}
    }
    # log to disk

    while (@log) {
	my $msg = pop @log;
	my $type = $msg->{type};
	my $logfile = $config{log}{logfile}{$type} || "exsite.log";
	&log($msg->{message},$logfile);
    }
}

sub log {
    my ($msg,$logfile) = @_;
    $logfile or $logfile = "exsite.log";
    if (-w $logfile) {
	$logfile =~ /^([-\w]+\.log)$/;
	my $secure_logfile = $1;
	if ($secure_logfile) {
	    if (open LOG, ">>$secure_logfile") {
		my ($time,$user,$op);
		$time = localtime;
		# don't use my_name(), as it will cause an infinite loop
		$user = exists $share{identity} ? $share{identity}{name} : "nobody";
		if (! $user) { $user = "unknown"; }
		$op = $ENV{REQUEST_URI};
		print LOG "$time\t$user\t$op\t$msg\n";
		close LOG;
		### FIXME: maintain an open logfile connection for better perf
	    }
	}
    }
}

# tell us if errors have occurred
# (can optionally specify a level and type)

sub OnError {
    my ($level,$type) = @_;
    $level or $level = 3;
    foreach my $msg (@{$share{diagnostics}}) {
	if ($type) {
	    return 1 if ($msg->{type} eq $type && $msg->{level} >= $level);
	}
	else {
	    return 1 if ($msg->{level} >= $level);
	}
    }
    return 0;
}

# Get the current diagnostic count
# $lastcount = returns the difference from the lastcount
# $level = counts only diagnostics of this level or higher
#
# To get the total error count from the start of the request:
# my $errcnt = &AtError();
#
# To test if any errors have appeared since the last check:
# if (&AtError($errcnt)) { # new errors occurred }

sub AtError {
    my ($lastcount,$level) = @_;
    $level or $level = 3;
    $lastcount or $lastcount = 0;
    if ($lastcount < 0) { $lastcount = 0; } # backwards-compat kludge
    my $count = 0;
    foreach my $msg (@{$share{diagnostics}}) {
	$count++ if ($msg->{level} >= $level);
    }
    return $count - $lastcount;
}

# warning: meaning of $type is not the same as ExSite::Base::fetch_diagnostics
sub fetch_diagnostics {
    my ($level,$type) = @_;
    my @message;
    $level or $level = 3;
    foreach my $msg (@{$share{diagnostics}}) {
	next if ($type && $type ne $msg->{type});
	next if ($level && $level > $msg->{level});
	push @message, $msg->{message};
    }
    return @message;
}

sub show_diagnostics {
    my ($format,$level,$type) = @_;
    my @diag = &fetch_diagnostics($level,$type);
    if (@diag > 0) {
	$format or $format = "HTML";
	if ($format =~ /HTML/i) {
	    my $ml = &get_obj("ML");
	    my $out;
	    foreach my $d (@diag) {
		$out .= $ml->div(join($ml->br, split(/\n/,$d)),{class=>"error"});
	    }
	    return $out ? $ml->div($out,{class=>"errors"}) : $out;
	}
	else {
	    return join("\n",@diag);
	}
    }
    return undef;
}

##################################################################
################ handlers                         ################
##################################################################

sub install_handlers {
    my $this = shift;
    my $db = $share{DB};
    if ($this == $db) {
	# standard CMS handlers here
	$db->handler("setup_querylib",\&CMS_setup_querylib);
	$db->handler("select_foreign_key",\&CMS_select_foreign_key);
    }
    # additional handlers unique to this installation 
    &my_handlers($this,@_);
}

# handler for foreign key selection on CMS tables
#
# - for *.template_id, select only valid templates
# - for *.section_id, select only from my_sites()
# - for page.parent_id, select only from other pages in same section
# - for content.page_id, select only from pages in same section
# - for section.parent_id, select only from other non-template sections

sub CMS_select_foreign_key {
    my ($this, $table, $rec, $col, $ftable) = @_;
    # ignore, if not using the CMS database
    if ($this->{db}{name} eq $config{server}{db}{name}) {
	if ($col eq "section_id") {
	    if (! $this->is_admin) {
		my @sites;
		@sites = $this->my_real_sites();
		return \@sites;
	    }
	    return undef;
	}
	if ($table eq "page" && $this->{action} !~ /search/) {
	    if ($col eq "template_id") {
		my @t;
		my $pagedata = $this->fetch("page",$rec);
		if ($pagedata) {
		    my $section = new ExSite::Section(id=>$pagedata->{section_id});
		    if ($pagedata->{type} eq "alias") {
			my $map = $section->get_mapped_pages(undef,{version=>$pagedata->{version}});
			@t = $map ? $map->collapse() : ();
		    }
		    else {
			@t = $section->get_templates();
		    }
		}
		else {
		    @t = $this->fetch_match("page",{type=>"template"});
		}
		return \@t;
	    }
	    elsif ($col eq "parent_id") {
		my (@possible_parents,$page,@parents);
		# select only from pages in same section
		$page = $this->fetch("page",$rec);
		@parents = $this->fetch_match("page",{section_id=>$page->{section_id}});
		# don't allow user to select a page as its own parent
		foreach my $parent (@parents) {
		    if ($parent->{page_id} ne $rec && 
			$parent->{type} =~ /page|alias|alternate/) {
			push @possible_parents, $parent;
		    }
		}
		return \@possible_parents;
	    }
	}
	if ($table eq "content" && $this->{action} !~ /search/) {
	    if ($col eq "page_id") {
		my ($content,$page);
		$content = $this->fetch("content",$rec);
		$page = $this->fetch("page",$content->{page_id});
		return $this->fetch_match("page",{section_id=>$page->{section_id}});
	    }
	}
	if ($table eq "section" && $this->{action} !~ /search/) {
	    if ($col eq "parent_id") {
		return $this->get_query("non-template sections",
					($rec =~ /^\d+$/) ? $rec : 0);
#		my $sql = "select * from section where type != 'website template'";
#		$sql .= ($rec =~ /^\d+$/) ? " and section_id != $rec" : "";
#		return $this->custom_query($sql);
	    }
	    elsif ($col eq "template_id") {
		return $this->fetch_match("page",{type=>"template"});
	    }
	}
    }
    return undef;
}

# The querylib is a library of SQL custom queries;  we prepopulate this with
# some useful queries used by the base CMS system.

sub CMS_setup_querylib {
    my $this = shift;
    # ignore, if not using the CMS database
    if ($this->{db}{name} eq $config{server}{db}{name}) {
	# used by foreign_key_handler, above
	$this->set_query("non-template sections",
			 ( sql => "select * from section where type != 'website template' and section_id != ?",
			   nparam => 1,
			   mode => "r",
			   keys => "section" )
			 );
	# load mappable pages
	$this->set_query("mapped pages",
			 ( sql => "select page.* from page where section_id=? and status != 'disabled' and type in ('page', 'alias') order by rank,page_id",
			   nparam => 1,
			   mode => "r",
			   keys => [ "page" ] )
			 );

	# load mappable alternates
	$this->set_query("mapped alternates",
			 ( sql => "select page.* from page where section_id=? and version=? and status != 'disabled' and type in ('alias', 'alternate') order by rank,page_id",
			   nparam => 2,
			   mode => "r",
			   keys => [ "page" ] )
			 );

	# prefetch content for a section
	$this->set_query("prefetch content",
			 ( sql => "select content.* from content,page where content.page_id=page.page_id and page.section_id=? and page.status=\"active\"",
			   nparam => 1,
			   mode => "r",
			   keys => [ "content", "page" ] )
			 );
	$this->set_query("prefetch content with mysql_timestamp_hack",
			 ( sql => "select content.content_id,content.page_id,content.name,content.allow_mime,content.description,content.type,content.mtime+0 as mtime,content.ctime+0 as ctime from content,page where content.page_id=page.page_id and page.section_id=? and page.status=\"active\"",
			   nparam => 1,
			   mode => "r",
			   keys => [ "content", "page" ] )
			 );
	# prefetch revisions for a section - NO CACHING
	$this->set_query("prefetch content_data",
			 ( sql => "select content_data.* from content_data,content,page where content_data.content_id=content.content_id and content.page_id=page.page_id and page.section_id=? and page.status=\"active\" order by content_data_id",
			   nparam => 1,
			   mode => "r", )
#			   keys => [ "content_data", "content", "page" ] )
			 );
	$this->set_query("prefetch content_data with mysql_timestamp_hack",
			 ( sql => "select content_data.content_data_id,content_data.content_id,content_data.mime_type,content_data.data,content_data.fdata,content_data.file,content_data.url,content_data.attributes,content_data.ctime+0 as ctime,content_data.note,content_data.thumb,content_data.thumb_attributes,content_data.ptime+0 as ptime where content_data.content_id=content.content_id and content.page_id=page.page_id and page.section_id=? and page.status=\"active\" order by content_data_id",
			   nparam => 1,
			   mode => "r", )
#			   keys => [ "content_data", "content", "page" ] )
			 );
	# same as previous, but does the file trimming in mysql instead of perl
	# this makes it cacheable
	$this->set_query("prefetch content_data with _fsize",
			 ( sql => "select r.content_data_id,r.content_id,r.mime_type,r.data,left(r.fdata,locate('#',r.fdata)) as fdata,r.attributes,left(r.thumb,locate('#',r.thumb)) as thumb,r.thumb_attributes,r.file,r.url,r.ctime,r.ptime,r.note,(length(r.fdata)-locate('#',r.fdata)) as _fsize,(length(r.thumb)-locate('#',r.thumb)) as _tsize from content_data as r,content,page where r.content_id=content.content_id and content.page_id=page.page_id and page.section_id=? and page.status=\"active\" order by content_data_id",
			   nparam => 1,
			   mode => "r",
			   keys => [ "content_data", "content", "page" ] )
			 );
	$this->set_query("prefetch content_data with _fsize and mysql_timestamp_hack",
			 ( sql => "select r.content_data_id,r.content_id,r.mime_type,r.data,left(r.fdata,locate('#',r.fdata)) as fdata,r.attributes,left(r.thumb,locate('#',r.thumb)) as thumb,r.thumb_attributes,r.file,r.url,r.ctime+0 as ctime,r.ptime+0 as ptime,r.note,(length(r.fdata)-locate('#',r.fdata)) as _fsize,(length(r.thumb)-locate('#',r.thumb)) as _tsize from content_data as r,content,page where r.content_id=content.content_id and content.page_id=page.page_id and page.section_id=? and page.status=\"active\" order by content_data_id",
			   nparam => 1,
			   mode => "r",
			   keys => [ "content_data", "content", "page" ] )
			 );
	# system messages in translation
	$this->set_query("system messages",
			 ( sql => "select sysmsg.*, sysmsg_alt.* from sysmsg, sysmsg_alt where sysmsg_alt.sysmsg_id=sysmsg.sysmsg_id and sysmsg_alt.version=?",
			   nparam => 1,
			   mode => "r",
			   keys => [ "sysmsg", "sysmsg_alt" ] )
			 );
	# services available to current user
	$this->set_query("member services",
			 ( sql => "select service.name from service,section,member_site_link where service.section_id=section.section_id and service.section_id=member_site_link.section_id and member_site_link.member_id=?",
			   nparam => 1,
			   mode => "r",
			   keys => [ "service", "section", "member_site_link" ] )
			 );
	# active oysters for a module in a section
	$this->set_query("oysters",
			 ( sql => "select page.*,content.content_id from page,content,content_data where page.page_id=content.page_id and content.content_id=content_data.content_id and page.section_id=? and content_data.mime_type=\"text/html\" and page.type=\"page\" and content_data.data rlike ? order by content_data.content_data_id desc",
			   nparam => 2,
			   mode => "r",
			   keys => [ "page","content","content_data"] )
			 );
#	# get all body objects in a section (current section only)
#	$this->set_query("section bodies",
#			 ( sql => "select content.* from page,content where page.page_id=content.page_id and page.type=\"library\" and page.status=\"active\" and content.name=\"body\" and page.section_id=?",
#			   nparam => 1,
#			   mode => "r",
#			   keys => [ "content", "page" ] )
#			 );
    }
    return undef;
}

#========================================
# DEPRECATED ADMIN SCREEN LAYOUT ROUTINES

sub page_header {
    my $out;
    if ($share{DB}) {
	$out = $share{DB}->run_handler("page_header",@_);
    }
    if (! $out) {
	$out = &my_page_header(@_);
	if (! $out) { 
	    # no custom header - use generic ExSite system wrapper
	    my ($title,$opt) = @_;
	    my $host = $config{server}{host} ? $config{server}{host}."." : $config{server}{host};
	    my $home = "$config{server}{protocol}://$host$config{server}{domain}$config{server}{HTMLpath}";
	    $out = "<HTML>
<HEAD>
<TITLE>$title</TITLE>
<meta http-equiv=\"Content-Type\" content=\"text/html; charset=$config{charset}\">
<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"$config{server}{HTMLpath}/_ExSite/css/ExSite.css\">
<script type='text/javascript' src='$config{server}{HTMLpath}/_ExSite/js/misc.js'></script>
</HEAD>
<BODY MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\" TOPMARGIN=\"0\" LEFTMARGIN=\"0\" RIGHTMARGIN=\"0\" BGCOLOR=\"#9999CC\">
<table cellspacing=0 cellpadding=0 border=0 bgcolor=#ccccdd width=100%>
<tr><td colspan=3 bgcolor=#eeeeff><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif height=1 width=1></td></tr>
<tr><td bgcolor=#eeeeff><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif height=1 width=1></td>
<td WIDTH=100%>

<!-- begin menubar -->

<table cellspacing=0 cellpadding=4 border=0 bgcolor=#ccccdd width=100%>

<td align=left WIDTH=100%>

<!-- logo -->
<a href=\"$home\"><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/exsite-button-round.gif height=30 width=30 border=0></a>

</td>
<td><A HREF=$config{server}{CGIpath}/$config{prog}{home}><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-home.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"ExSite Home\"></A></td>
<td><A HREF=\"javascript:popup_medium('$config{server}{CGIpath}/$config{prog}{help}')\"><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-help.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"Help\"></A></td>
<td><A HREF=$config{server}{CGIpath}/$config{prog}{logout}><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-exit.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"Log Out\"></A></td>
</tr>
</table>

<!-- end menubar -->

</td>
<td bgcolor=#333366><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/null.gif height=1 width=1></td>
</tr>
<tr><td colspan=6 bgcolor=#333366><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/null.gif height=1 width=1></td></tr>
</table><br>\n\n";

	    if (! $opt || ref $opt) {
		# opt flags the webtop in the default presentation
		# if !opt, then draw a window
		$out .= "<div align=center>
    <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0 WIDTH=90% CLASS=framed>
      <TR>
        <TD CLASS=titlebar><h1>$title</h1></TD>
      </TR>
      <TR>
        <TD BGCOLOR=#FFFFFF><DIV CLASS=htmlbox>\n";
	    }
	    elsif ($opt != 1) {
		# opt is actual content, not a flag
		$out .= $opt;
	    }
	}
    }
    
    $out .= "<p class='error'>Warning: this is a deprecated page layout.</p>\n";
    # do not modify anything below this point
    
    if ($config{page_buffering} || $share{PageBuf}) { 
	return $out; 
    }
    else { 
	print "Content-type: text/html\n\n";
	print $out; 
    }
}

sub page_footer {
    my $out;
    if ($share{DB}) {
	$out = $share{DB}->run_handler("page_footer",@_);
    }
    if (! $out) {
	$out = &my_page_footer(@_);
	if (! $out) { 
	    # no special page footer - use ExSite system wrapper
	    my ($ref) = @_;
	    if (! $ref || ref $ref) {
		# close the pseudo-window
		$out = "        </DIV></TD>\n      </TR>\n    </TABLE>\n</div>\n";
	    }
	    elsif ($ref != 1) {
		# custom footer
		$out = $ref;
	    }
	    if ($share{DB}) {
		# there is a database object - check for any diagnostics to report
		my $err;
		if ($share{DB}->errorcheck()) {
		    $err .= "<h3>Errors:</h3>\n";
		    $err .= $share{DB}->show_diagnostics("error","html");
		}
		if ($share{DB}->errorcheck("warn")) {
		    $err .= "<h3>Warnings:</h3>\n";
		    $err .= $share{DB}->show_diagnostics("warn","html");
		}
		if ($err) {
		    $out = "<div class=\"error\">\n$err\n</div>\n";
		}
	    }
	    $out .= "</BODY></HTML>\n";
	}
    }
    if ($config{page_buffering} || $share{PageBuf}) { return $out; }
    else { print $out; }
}

1;

=pod

=head1 ExSite::Config - ExSite configuration and initialization

C<Config.pm> contains general configuration parameters,
site-specific setup data, and global variables used throughout
the ExSite system.

This file is designed so that it never has to be modified.  To set
configuration variables, use an "exsite.conf" file, which will be
loaded automatically, overriding the presets below.  This file can be
placed in C<cgi-bin/>, or C<cgi-bin/conf/>.  Special configuration 
routines and functions can be defined in the module C<myConfig.pm>.

=head2 Usage:

C<use ExSite::Config;>

This is required in all ExSite programs.  In addition, every program
should also include the following first line of executable code:

C<&exsite_init;>

=head2 Configuration Management

The system defines six global configuration hashes, described below.
These are exported to all programs that use C<ExSite::Config>.

=over 4

=item C<%config>

C<%config> is used for system constants that do not normally change.

The system constants in C<%config> are read first from the hard-coded
values in Config.pm, and then from the local configuration file 
"exsite.conf", which can override the former values.  
Modifications to the site configuration should be made 
in exsite.conf, not in Config.pm.

Specific configuration parameters are described in the ExSite
Configuration documentation, or in Config.pm.

Any module or other package can read its own configuration file(s) into
C<%config>, by calling: C<read_exsite_conf()>.

In a plain CGI setup, C<%config> is re-read on each request, so
configuration changes are picked up immediately.  In a persistent perl
setup, C<%config> will remain defined from the previous request, and
you will need to restart your persistent perl to refresh C<%config>.
In a setup with the persistent data store (see C<%store>, below),
ExSite will try to load a cached version of C<%config> from C<%store>
before reparsing the raw config files.  To change the configuration,
you will need to reset the C<%store> to force the data to reload.

=item C<%share>

C<%share> is used as a general-purpose shared-memory area for the current
request.  C<%share> is cleared on each page request.

The C<%share> hash is used for sharing data between widely separated
modules and routines.  The primary difference between C<%config> and
C<%share> is that the former is regarded as persistent (values stay
the same from request to request), while the latter is volatile
(values vary from request to request).

The differences are not especially significant for normal CGI setups,
since C<%config> must be reloaded on each request anyway.  But in a
persistent perl or persistent data store configuration, C<%config> may
only be configured once, whereas C<%share> will be cleared and rebuilt
on every request.  (If you do not follow this convention, you may have
problems when switching to use persistence.)

C<%share> is cleared every time C<&exsite_init()> is called.
C<%share> will be used to store pointers to the current database and
page objects (C<$share{DB}> and C<$share{Page}>, respectively), plus
any other shareable data placed there by localization routines and
content modules.

=item C<%session> 

C<%session> stores semi-persistent data for a particular user.  That
is, it remembers data between different page requests, but clears the
data after a certain amount of time has passed with no activity.  The
C<%session> hash only holds data for the current user; other sessions'
data are not available.  This is a useful mechanism for temporarily
preserving a visitor's state.  Sessions last for 1 hour by default,
but are automatically renewed if used.  In other words, if no activity
is detected under a session for one hour, the session will be
terminated.

Session management is automatically enabled if you are using the
persistent data store (C<%store>).  Actual session records are created
only if data is written to C<%session>, however.  That is, there is no
persistent record of a session if nothing is ever written into
C<%session>.  The number of session records indicates the number of
users who have session data that is being tracked, not necessarily the
total number of users active on the site.

C<%session> will will automatically be populated with data previously
written to that user's session.  (See C<ExSite::Session> for technical
details.)  These data include C<_mtime> and C<_ctime>, which are
timestamp values (Unix system times) of the last modification of the
session data, and the creation time of the session.

B<Warnings:>
If persistent storage is not enabled, sessions are not especially
useful, as they will not remember data between requests.  However,
there is no harm in reading or writing to C<%session> in this case, and
it will behave as a normal hash variable.

(You can also use C<%cookie> to store semi-persistent data.
C<%session> is more secure, powerful, and reliable if enabled, since
the data is stored on the server, and can store complex structures.
Note that when a session is initiated, the session ID is stored in
C<%cookie>, so both are in fact needed to make proper use of
sessions.)

=item C<%cookie> 

C<%cookie> is used for session-like variables that are tracked
client-side rather than server-side.  As the name implies, these
variables are stored as cookies in the user's browser.  The C<%cookie>
hash automatically acquires all of the relevant cookies, and can be
inspected to view the values of those cookies at any time.

Changing/setting cookie values is also easy; simply assign a new
value to a cookie name, eg.

    $cookie{name} = $value;

The appropriate HTTP headers will be printed to ensure that the new
cookie data is sent to the client browser for storage.

B<Warnings:>
Because cookies are stored as simple text strings on the client browser,
cookie values should also be simple text strings.  Do not store 
references or objects in cookies.

Because browsers can disregard cookies, and cookies can be manually
deleted or edited at any time, there is no guarantee that cookies will
in fact persist, nor that they will not be tampered with.  They are
also transmitted in cleartext, so are vulnerable to snooping.  That
means that C<%cookie> is a useful convenience, but not a secure place
for storing data.  Sensitive data should be encrypted before being
placed into a cookie.  For example:

=over 4

=item ExSite identity cookie

In "cookie" mode, it stores DES-encrypted password to validate the user 
identity.  In "crypt" mode, it stores TEA-encrypted username and access
level.

=item ExSite session ID

The session ID is a 16-character random text string.  This
makes active session IDs hard to guess, and fairly secure from a
keyspace attack.  However, it is not very secure from a spoofing
attack if an active session ID is sniffed from network traffic. 
Use https/SSL to guard against that.

=back

=item C<%store>

The persistent data store is a place to store globally persistent
dynamic data.  While C<%config> stores parameters that do not change,
C<%store> stores data that is regularly modified but needs to be
available to future requests.  C<%store> is the underlying mechanism
behind sessions, database caches, and configuration caches.

The persistent data store is optional; by default it is not enabled,
and C<%store> is just a normal hash whose values are cleared on each
request.  (Which means that sessions and caches do not store data
beyond the current request.)  If the persistent store is enabled (see
the docs for ExSite::Store) you automatically get the benefit of file
caching, database caching, and sessions.  You do not need to interact
directly with C<%store> to gain these features; ExSite manages that on
your behalf.

Items can be placed in the store with an expiry time (which can be 0
for no limit).  Items past their expiry time are automatically purged
from the store.

=item C<%msg>

The C<%msg> hash is a simple Internationalization facility.  By default
it simply echoes its key back as a value.  If you have a language other
than English defined as your default language, however, then it will
attempt to return an appropriate translation of the key instead.  By this
means, a line such as:

    print $msg{'Hello World'};

will print out the "Hello World" message in whatever language the 
system is defined to use (assuming that an appropriate dictionary 
is installed).

=back

=head1 Setup and Configuration Routines

=head2 server_autoconf()

This routine attempts to set the various server configuration
parameters (ie. file paths and machine names) automatically, by
inferring them from the server environment.  It works for simple
system configurations, but may get confused by unusual server
setups.  In the latter case, disable autoconfiguration by setting
C<server.auto = 0> in your configuration file.  You will then need
to specify the various server parameters manually.

B<Warning:> auto-configuration may fail when testing CGI programs from
the command line for testing/debugging purposes, because it may infer
different paths than the web server uses.  If you need to have correct
configuration, then either hard-code your configurations in
exsite.conf, or set some environment variables on your command line to
resemble the server environment.  The following environment variables
are used for auto-configuration: DOCUMENT_ROOT, SCRIPT_NAME,
SCRIPT_FILENAME, HTTPS, HTTP_HOST.

=head2 read_exsite_conf()

This routine loads configuration values from files.  By default 
it looks for "exsite.conf" files in the conf, ExSite and . directories.
An alternate filename can be given.  By default configuration values 
are read straight into the root of C<%config>, but a starting key can 
be given, eg. C<&read_exsite_conf("MyModule.conf","MyModule")> will
read the values in MyModule.conf into C<$config{MyModule}{...}>.

Configuration files contain parameters and values, separated by "=" 
(with any amount of whitespace around the "=".  Values extend from the 
first non-whitespace character after the "=", to the end of line.
Simple parameters are placed directly into C<%config>'s keys.  
Parameters of the form "a.b.c" are stored as C<$config{a}{b}{c}>. 
Config array references can be created using "+=" to set the values
in the array.

Example exsite.conf settings:

    trim_whitespace = 0
    form.style=list
    db_ops.new.url = append.cgi

=head2 exsite_init()

This routine configures ExSite, and performs any local initializations.
This routine is called automatically at the start of all
generic CGI scripts, and should be called by custom scripts
as well.

Local initialization can be done in C<&myConfig::my_exsite_init>,
which is called automatically from here.

=head2 exsite_close()

Call this at the end of every ExSite program to ensure that 
files, caches, sessions, etc. are properly flushed.

=head2 get_obj()

There are a handful of global objects that ExSite commonly works with.
The most common are the primary database object, and the current page
being constructed.  These are normally stored in %share to make them 
available to all components.

You can use get_obj() to fetch a predefined object safely.  That is,
it will return the predefined object if it exists, and it will create
the object if it does not exist.

Example:

C<my $db = &get_obj('DB');>

=head2 install_handlers()

Handlers allow plug ins or other subsystems to register special 
business logic that should take precedence over the default logic
in certain cases.

The ExSite base system has to do this in a few cases for its content
management tables, which have their own logic that differs from the
generic database handling.  Otherwise, it is expected that most 
handlers will be site-specific (registered in myConfig::my_handlers() )
or plug-in-specific (registered from the plug-in module in 
Modules/).

=cut

