#!/usr/bin/perl
#!/usr/bin/speedy
#-----------------------------------------------------------------------
# page.cgi - generate and display a web page
#
# USAGE:
#    page.cgi?_id=xxx
#    page.cgi/path/to/file.html
#
# Parameters:
#    _ = aggregated parameters (encrypted)
#    _id = id number of page to display
#    _cms = inlined editing flag (enable editing if true)
#    _cmsmode = persistent inlined editing flag (enable editing for all pages)
#    (other) = parameters for plug-in modules
#
# Other parameters are passed through to dynamic content objects that
# are embedded in the page.  By convention, parameters directed at page.cgi
# have a preceding underscore;  other parameters are intended for embedded
# dynamic content modules.
#-----------------------------------------------------------------------
#
#   Copyright 2001-2007 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Page;
use ExSite::Session;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Input;

use vars qw($utime1 $stime1 $cutime1 $cstime1 $redirect);
($utime1,$stime1,$cutime1,$cstime1) = times;
&exsite_init;

if ($config{kill}) {
    # gaah!  We've been killed!
    &kill_page();
}
elsif ($config{busy} || $config{throttle}{user}{rate} || $config{throttle}{system}{rate}) {
    # rate limiters are enabled
    if ($ENV{REQUEST_METHOD} ne "GET") {
	# only rate-limit GET requests (eg. crawlers)
	&real_page();
    }
    elsif ($config{priority_page} && $ENV{REQUEST_URI} =~ /$config{priority_page}/) {
	# priority pages are not rate-limited
	&real_page();
    }
    elsif ($config{busy} && (&sysload() > $config{busyload})) {
	# system is swamped; delay the page
	&busy_page();
    }
    elsif (($config{throttle}{user}{rate} || $config{throttle}{system}{rate}) && 
	   &throttle()) {
	# someone is making too many requests; delay the page
	&throttle_page();
    }
    else {
	# all is calm; serve the page
	&real_page();
    }
}
else {
    # normal operation
    &real_page();
}

&exsite_close;
print &debug_status($utime1,$stime1,$cutime1,$cstime1) if (! $redirect && $config{debug} > 1);

# crude benchmarking tool for ExSite developers
# This measures pure code execution time, not Perl and library overhead.

# The kill page is displayed if the killswitch is tripped.  You can use this
# feature when performing system maintenance, or for disabling the
# most resource-intensive parts of a site for overload handling.

sub kill_page {
    my $killurl = $config{killurl} || "$config{server}{HTMLpath}/_ExSite/notavailable.html";
    print "Location: $killurl\n\n";
    $redirect = 1;
    return;
}

# The busy page is displayed if you have the busy switch enabled, and the
# 1 minute load average exceeds a certain threshold.  This serves primarily
# to limit the impact of DOS attacks and bad robots, but gives real readers
# a chance to reload.

sub busy_page {
    my $busyurl = $config{busyurl} || "$config{server}{HTMLpath}/_ExSite/serverbusy.html";
    print "Location: $busyurl\n\n";
    $redirect = 1;
    return;
}

# busy() decides if we are too busy to serve the real page.  This is the
# case if our 1 minute load average exceeds the value of $config{busyload}

sub busy {
    return (&sysload() > $config{busyload});
}

# The throttle page is displayed if you have the traffic throttler enabled, 
# and the request rate exceeds the allowed rate.  It enforces a wait until the
# request rate drops to an acceptable level.

sub throttle_page {
    my $busyurl = $config{throttle}{url} || "$config{server}{HTMLpath}/_ExSite/throttle.html";
    print "Location: $busyurl\n\n";
    $redirect = 1;
    return;
}

# show the real page...

sub real_page {
    my $db = new ExSite::Form;
    my $p = new ExSite::Page;

    # check for shortened URL
    if ($config{token}{short_url} && $ENV{PATH_INFO} =~ /^\/\w+/) {
	# might be a token
	my $name = $ENV{PATH_INFO};
	$name =~ s/^\///;
	my $t = tied %token;
	if ($t && $t->enabled) {
	    my $url = $token{$name};
	    if ($url && $url =~ /^(([^:\/?\#]+):)?(\/\/([^\/?\#]*))?([^?\#]*)(\?([^\#]*))?(\#(.*))?/) {
		# yep, we have a shortened URL
		my $uri = new ExSite::URI(uri=>$url);
		my $script_name = $uri->get("script_name");
		if ($script_name =~ /\/$config{prog}{page}$/) {
		    # shortened page url
		    $ENV{SCRIPT_NAME} = $script_name;
		    $ENV{QUERY_STRING} = $uri->get("query");
		    $ENV{PATH_INFO} = $uri->get("path_info");
		    my $uid = $t->uid($name);
		    if ($uid) {
			my $user = $db->fetch($config{auth}{user_table},$uid);
			if (ref $user eq "HASH") {
			    # do a temporary login instead?
			    $db->do_login($user);
			}
		    }
		}
		else {
		    # non-page URL, must use a redirect
		    my $uid = $t->uid($name);
		    if ($uid) {
			my $url = $uri->authorize_url($uid,1);
			$p->redirect($url);
			$p->show();
			$redirect = 1;
			return;
		    }
		}
	    }
	}
    }

    my $in = new ExSite::Input;
    my %query = %{$in->query};
    my $html;

    # check to see if we need embedded CMS features in this page;

    my $access = $db->authorize;
    my $cmsflag = ($access >= 2 && exists $query{_cms}) ? $query{_cms} : undef;

    # generate the page

    my $got_page_id = 0;
    my %pagesetup = ( 
	cms=>$cmsflag, 
	preload=>($config{page}{dyn_publish_method} ne "template" && ! defined $cmsflag),
	);
    if ($query{_id}) {
	# /cgi/page.cgi?_id=XXX
	$pagesetup{id} = $query{_id};
	$got_page_id = 1;
    }
    elsif ($ENV{PATH_INFO}) {
	# /cgi/page.cgi/section/page.html
	$pagesetup{path} = 1;
	$got_page_id = 1;
    }
    if ($got_page_id) {
	# save page pointer for benefit of DCDs
	$share{Page} = $p;
	$p->setup( %pagesetup );
	$p->fix_path();   # in case we came in from the wrong URL type
	
	if ($p->has_content()) {
	    # looks like we set up a real page

	    # set the version/language
	    my $version = $p->get_my("version");
	    if ($version) {
		# try to use matching system messages
		(tied %msg)->version($version);
	    }
		
	    if ($p->authorize()) {
		if ($p->is_publishable && 
		    $share{DB}->level()==0 && 
		    $p->is_default_view()) {
		    # member of public trying to view static page dynamically
		    $p->redirect($p->get_url_static());
		}
		else {
		    # insert all the content
		    my $mode;
		    if ($p->is_template && defined $cmsflag) {
			$mode = "dynamic";
		    }
		    $p->expand(cms=>$cmsflag,mode=>$mode);
		    if ($cmsflag) {
			$p->header("X-XSS-Protection: 0");
		    }
		}
	    }
	    else {
		# permission denied!
		if (! $p->is_active) {
		    # disabled
		    &no_page($p,"Sorry, the page you have requested is not available at this time.");
		}
		elsif (exists $p->{is_valid} && ! $p->{is_valid}) {
		    # cross-domain page lookup - only allowed for admins
		    &no_page($p,"Invalid page.");
		}
		else {
		    # Require user to log in before proceeding.
		    # display any permission denied errors that may exist
		    my $login_form = $p->show_diagnostics("error","html");
		    my $goto = $config{server}{server}.$ENV{REQUEST_URI};
		    $login_form .= $share{DB}->login_form("",$goto);
		    $p->errorpage($login_form);
		}
	    }
	}
	else {
	    my $err = $msg{"Sorry, cannot build this page. (It appears to lack a template.)"};
	    $p->errorpage($p->error($err));
	}
    }
    else {
	&no_page($p,"No page to display!");
    }
    if ($in->path()) {
	# unused path info
	my $errortype = $config{page}{unused_path_error};
	if ($errortype) {
	    my $err = $msg{"Unknown resource:"}." ".$in->path();
	    if ($errortype eq "fatal") {
		$p->errorpage($p->error($err));
	    }
	    elsif ($errortype eq "error") {
		$p->error($err);
	    }
	    elsif ($errortype eq "warn") {
		$p->warn($err);
	    }
	    elsif ($errortype eq "info") {
		$p->info($err);
	    }
	    elsif ($errortype eq "404") {
		$p->error404();
	    }
	}
    }
    $p->show;
}

sub no_page {
    my ($p,$err) = @_;
    $config{page}{no_page_404} ?
	$p->error404() :
	$p->errorpage($p->error($msg{$err}));
    return;
}
